// Copyright 2010 the V8 project authors. All rights reserved.
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
//       copyright notice, this list of conditions and the following
//       disclaimer in the documentation and/or other materials provided
//       with the distribution.
//     * Neither the name of Google Inc. nor the names of its
//       contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// Test constant folding of smi operations that overflow a 32-bit int
// See http://code.google.com/p/v8/issues/detail?id=580

function num_ops() {
  var x;
  var tmp = 0;
  x = (tmp = 1578221999, tmp)+(tmp = 572285336, tmp);
  assertEquals(2150507335, x, "++");
  x = 1578221999 + 572285336;
  assertEquals(2150507335, x);

  x = (tmp = -1500000000, tmp)+(tmp = -2000000000, tmp);
  assertEquals(-3500000000, x, "+-");
  x = -1500000000 + -2000000000;
  assertEquals(-3500000000, x);

  x = (tmp = 1578221999, tmp)-(tmp = -572285336, tmp);
  assertEquals(2150507335, x, "--");
  x = 1578221999 - -572285336;
  assertEquals(2150507335, x);

  x = (tmp = -1500000000, tmp)-(tmp = 2000000000, tmp);
  assertEquals(-3500000000, x, "-+");
  x = -1500000000 - 2000000000;
  assertEquals(-3500000000, x);
}

num_ops();
